from histomicstk.cli import utils as cli_utils
from histomicstk.segmentation import label as htk_label

from .compute_fsd_features import compute_fsd_features
from .compute_gradient_features import compute_gradient_features
from .compute_haralick_features import compute_haralick_features
from .compute_intensity_features import compute_intensity_features
from .compute_morphometry_features import compute_morphometry_features


def compute_nuclei_features(im_label, im_nuclei=None, im_cytoplasm=None,
                            fsd_bnd_pts=128, fsd_freq_bins=6, cyto_width=8,
                            num_glcm_levels=32,
                            morphometry_features_flag=True,
                            fsd_features_flag=True,
                            intensity_features_flag=True,
                            gradient_features_flag=True,
                            haralick_features_flag=True,
                            tile_info=None,
                            im_nuclei_seg_mask=None,
                            format=None,
                            return_nuclei_annotation=False,
                            ):
    """
    Calculates features for nuclei classification

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.

    im_nuclei : array_like
        Nucleus channel intensity image.

    im_cytoplasm : array_like
        Cytoplasm channel intensity image.

    fsd_bnd_pts : int, optional
        Number of points for boundary resampling to calculate fourier
        descriptors. Default value = 128.

    fsd_freq_bins : int, optional
        Number of frequency bins for calculating FSDs. Default value = 6.

    cyto_width : float, optional
        Estimated width of the ring-like neighborhood region around each
        nucleus to be considered as its cytoplasm. Default value = 8.

    num_glcm_levels: int, optional
        An integer specifying the number of gray levels For example, if
        `NumLevels` is 32,  the intensity values of the input image are
        scaled so they are integers between 0 and 31.  The number of gray
        levels determines the size of the gray-level co-occurrence matrix.

        Default: 32

    morphometry_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        morphometry (size and shape) features.
        See histomicstk.features.compute_morphometry_features for more details.

    fsd_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        Fouried shape descriptor (FSD) features.
        See `histomicstk.features.compute_fsd_features` for more details.

    intensity_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        intensity features from the nucleus and cytoplasm channels.
        See `histomicstk.features.compute_fsd_features` for more details.

    gradient_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        gradient/edge features from intensity and cytoplasm channels.
        See `histomicstk.features.compute_gradient_features` for more details.

    haralick_features_flag : bool, optional
        A flag that can be used to specify whether or not to compute
        haralick features from intensity and cytoplasm channels.
        See `histomicstk.features.compute_haralick_features` for more details.

    return_nuclei_annotation :  bool, optional
        Returns the nuclei annotation if kept True

    Returns
    -------
    fdata : pandas.DataFrame
        A pandas data frame containing the features listed below for each
        object/label
    nuclei_annot_list : List
        List containing the boundaries of segmented nuclei in the input image.

    Notes
    -----
    List of features computed by this function

    Identifier
        Location of the nucleus and its code in the input labeled mask.
        Columns are prefixed by *Identifier.*. These include ...

        Identifier.Label (int) - nucleus label in the input labeled mask

        Identifier.Xmin (int) - Left bound

        Identifier.Ymin (int) - Upper bound

        Identifier.Xmax (int) - Right bound

        Identifier.Ymax (int) - Lower bound

        Identifier.CentroidX (float) - X centroid (columns)

        Identifier.CentroidY (float) - Y centroid (rows)

        Identifier.WeightedCentroidX (float) - intensity-weighted X centroid

        Identifier.WeightedCentroidY (float) - intensity-weighted Y centroid

    Morphometry (size, shape, and orientation) features of the nuclei
        See histomicstk.features.compute_morphometry_features for more details.
        Feature names prefixed by *Size.*, *Shape.*, or *Orientation.*.

    Fourier shape descriptor features
        See `histomicstk.features.compute_fsd_features` for more details.
        Feature names are prefixed by *FSD*.

    Intensity features for the nucleus and cytoplasm channels
        See `histomicstk.features.compute_fsd_features` for more details.
        Feature names are prefixed by *Nucleus.Intensity.* for nucleus features
        and *Cytoplasm.Intensity.* for cytoplasm features.

    Gradient/edge features for the nucleus and cytoplasm channels
        See `histomicstk.features.compute_gradient_features` for more details.
        Feature names are prefixed by *Nucleus.Gradient.* for nucleus features
        and *Cytoplasm.Gradient.* for cytoplasm features.

    Haralick features for the nucleus and cytoplasm channels
        See `histomicstk.features.compute_haralick_features` for more details.
        Feature names are prefixed by *Nucleus.Haralick.* for nucleus features
        and *Cytoplasm.Haralick.* for cytoplasm features.

    See Also
    --------
    histomicstk.features.compute_morphometry_features,
    histomicstk.features.compute_fsd_features,
    histomicstk.features.compute_intensity_features,
    histomicstk.features.compute_gradient_features,
    histomicstk.features.compute_haralick_features

    """
    import pandas as pd
    from skimage.measure import regionprops

    # sanity checks
    if any([
        intensity_features_flag,
        gradient_features_flag,
        haralick_features_flag,
    ]):
        assert im_nuclei is not None, 'You must provide nuclei intensity!'

    # TODO: this pipeline uses loops a lot. For each set of features it
    #  iterates over all nuclei, which may become an issue when one needs to
    #  do this for lots and lots of slides and 10^6+ nuclei. Consider
    #  improving efficiency in the future somehow (cython? reuse? etc)

    feature_list = []

    # get the objects in im_label
    nuclei_props = regionprops(im_label, intensity_image=im_nuclei)

    # extract object locations and identifiers
    idata = pd.DataFrame()
    for i, nprop in enumerate(nuclei_props):
        idata.at[i, 'Label'] = nprop.label
        idata.at[i, 'Identifier.Xmin'] = nprop.bbox[1]
        idata.at[i, 'Identifier.Ymin'] = nprop.bbox[0]
        idata.at[i, 'Identifier.Xmax'] = nprop.bbox[3]
        idata.at[i, 'Identifier.Ymax'] = nprop.bbox[2]
        idata.at[i, 'Identifier.CentroidX'] = nprop.centroid[1]
        idata.at[i, 'Identifier.CentroidY'] = nprop.centroid[0]
        if im_nuclei is not None:
            # intensity-weighted centroid
            wcy, wcx = nprop.weighted_centroid
            idata.at[i, 'Identifier.WeightedCentroidX'] = wcx
            idata.at[i, 'Identifier.WeightedCentroidY'] = wcy
    feature_list.append(idata)

    # compute cytoplasm mask
    if im_cytoplasm is not None:

        cyto_mask = htk_label.dilate_xor(im_label, neigh_width=cyto_width)

        cyto_props = regionprops(cyto_mask, intensity_image=im_cytoplasm)

        # ensure that cytoplasm props order corresponds to the nuclei
        lablocs = {v['label']: i for i, v in enumerate(cyto_props)}
        cyto_props = [cyto_props[lablocs[v['label']]] if v['label'] in lablocs else None
                      for v in nuclei_props]

    # compute morphometry features
    if morphometry_features_flag:

        fmorph = compute_morphometry_features(im_label, rprops=nuclei_props)

        feature_list.append(fmorph)

    # compute FSD features
    if fsd_features_flag:

        ffsd = compute_fsd_features(im_label, fsd_bnd_pts, fsd_freq_bins,
                                    cyto_width, rprops=nuclei_props)

        feature_list.append(ffsd)

    # compute nuclei intensity features
    if intensity_features_flag:

        fint_nuclei = compute_intensity_features(im_label, im_nuclei,
                                                 rprops=nuclei_props)
        fint_nuclei.columns = ['Nucleus.' + col
                               for col in fint_nuclei.columns]

        feature_list.append(fint_nuclei)

    # compute cytoplasm intensity features
    if intensity_features_flag and im_cytoplasm is not None:

        fint_cytoplasm = compute_intensity_features(cyto_mask, im_cytoplasm,
                                                    rprops=cyto_props)
        fint_cytoplasm.columns = ['Cytoplasm.' + col
                                  for col in fint_cytoplasm.columns]

        feature_list.append(fint_cytoplasm)

    # compute nuclei gradient features
    if gradient_features_flag:

        fgrad_nuclei = compute_gradient_features(im_label, im_nuclei,
                                                 rprops=nuclei_props)
        fgrad_nuclei.columns = ['Nucleus.' + col
                                for col in fgrad_nuclei.columns]

        feature_list.append(fgrad_nuclei)

    # compute cytoplasm gradient features
    if gradient_features_flag and im_cytoplasm is not None:

        fgrad_cytoplasm = compute_gradient_features(cyto_mask, im_cytoplasm,
                                                    rprops=cyto_props)
        fgrad_cytoplasm.columns = ['Cytoplasm.' + col
                                   for col in fgrad_cytoplasm.columns]

        feature_list.append(fgrad_cytoplasm)

    # compute nuclei haralick features
    if haralick_features_flag:

        fharalick_nuclei = compute_haralick_features(
            im_label, im_nuclei,
            num_levels=num_glcm_levels,
            rprops=nuclei_props,
        )

        fharalick_nuclei.columns = ['Nucleus.' + col
                                    for col in fharalick_nuclei.columns]

        feature_list.append(fharalick_nuclei)

    # compute cytoplasm haralick features
    if haralick_features_flag and im_cytoplasm is not None:

        fharalick_cytoplasm = compute_haralick_features(
            cyto_mask, im_cytoplasm,
            num_levels=num_glcm_levels,
            rprops=cyto_props,
        )

        fharalick_cytoplasm.columns = ['Cytoplasm.' + col
                                       for col in fharalick_cytoplasm.columns]

        feature_list.append(fharalick_cytoplasm)

    # Merge all features
    fdata = pd.concat(feature_list, axis=1)

    if return_nuclei_annotation:
        # Create nuclei segmentation with the generated regionprops
        nuclei_annot_list, selected_rows = cli_utils.create_tile_nuclei_annotations(
            im_nuclei_seg_mask, tile_info, format, nuclei_props)

        # Drop all rows which are not found in nuclei detection
        fdata = fdata[fdata.index.isin(selected_rows)]
        return fdata, nuclei_annot_list

    return fdata



"""
Created on Fri Aug 23 17:14:38 2019.
@author: tageldim
"""
import json
import os
import sys

import pytest
from pandas import read_csv

from histomicstk.annotations_and_masks.annotation_and_mask_utils import (
    delete_annotations_in_slide, parse_slide_annotations_into_tables)
from histomicstk.annotations_and_masks.masks_to_annotations_handler import (
    _discard_nonenclosed_background_group,
    get_annotation_documents_from_contours)
from histomicstk.annotations_and_masks.polygon_merger import Polygon_merger
from histomicstk.annotations_and_masks.polygon_merger_v2 import \
    Polygon_merger_v2

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../'))
from tests.datastore import datastore  # noqa
from tests.htk_test_utilities import getTestFilePath, girderClient  # noqa


class TestPolygonMerger:
    """Test methods for polygon merger."""

    @pytest.mark.usefixtures('girderClient')  # noqa
    def test_polygon_merger_tiled_masks(self, girderClient):  # noqa
        """Test Polygon_merger.run()."""
        original_iteminfo = girderClient.get(
            '/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

        folder = girderClient.post(
            '/folder', data={
                'parentId': original_iteminfo['folderId'],
                'name': 'test-polygon-merge',
            })

        # copy the item
        sampleSlideItem = girderClient.post(
            '/item/%s/copy' % original_iteminfo['_id'], data={
                'name': 'TCGA-A2-A0YE-01Z.svs',
                'copyAnnotations': True,
                'folderId': folder['_id'],
            })

        # read GTCodes dataframe
        gtcodePath = getTestFilePath('sample_GTcodes.csv')
        GTCodes_df = read_csv(gtcodePath)
        GTCodes_df.index = GTCodes_df.loc[:, 'group']

        # This is where masks for adjacent rois are saved
        mask_loadpath = getTestFilePath(os.path.join(
            'annotations_and_masks', 'polygon_merger_roi_masks'))
        maskpaths = [
            os.path.join(mask_loadpath, j) for j in os.listdir(mask_loadpath)
            if j.endswith('.png')]

        # init and run polygon merger on masks grid
        pm = Polygon_merger(
            maskpaths=maskpaths, GTCodes_df=GTCodes_df,
            discard_nonenclosed_background=True, verbose=1)
        contours_df = pm.run()

        # make sure it is what we expect
        assert contours_df.shape == (13, 13)
        assert set(contours_df.loc[:, 'group']) == {
            'roi', 'mostly_tumor', 'mostly_stroma',
            'mostly_lymphocytic_infiltrate'}

        # deleting existing annotations in target slide (if any)
        sampleSlideId = str(sampleSlideItem['_id'])
        delete_annotations_in_slide(girderClient, sampleSlideId)

        # get list of annotation documents
        annotation_docs = get_annotation_documents_from_contours(
            contours_df.copy(), separate_docs_by_group=True,
            docnamePrefix='test',
            verbose=False, monitorPrefix=sampleSlideId + ': annotation docs')

        # post annotations to slide -- make sure it posts without errors
        for annotation_doc in annotation_docs:
            resp = girderClient.post(
                '/annotation?itemId=' + sampleSlideId, json=annotation_doc)
            assert 'annotation' in resp.keys()

    def test_polygon_merger_rtree(self):
        """Test Polygon_merger_v2.run()."""
        annotationPath = datastore.fetch(
            'TCGA-A2-A0YE-01Z-00-DX1_GET_MergePolygons.svs_annotations.json')  # noqa
        slide_annotations = json.load(open(annotationPath))
        _, contours_df = parse_slide_annotations_into_tables(slide_annotations)

        # init & run polygon merger
        pm = Polygon_merger_v2(contours_df, verbose=0)
        pm.unique_groups.remove('roi')
        pm.run()

        # make sure it is what we expect
        assert pm.new_contours.shape in [(16, 16), (15, 16)]
        assert set(pm.new_contours.loc[:, 'group']) == {
            'mostly_tumor', 'mostly_stroma', 'mostly_lymphocytic_infiltrate'}

        # add colors (aesthetic)
        for group in pm.unique_groups:
            cs = contours_df.loc[contours_df.loc[:, 'group'] == group, 'color']
            pm.new_contours.loc[
                pm.new_contours.loc[:, 'group'] == group, 'color'] = cs.iloc[0]

        # get rid of nonenclosed stroma (aesthetic)
        pm.new_contours = _discard_nonenclosed_background_group(
            pm.new_contours, background_group='mostly_stroma')

        # get list of annotation documents
        annotation_docs = get_annotation_documents_from_contours(
            pm.new_contours.copy(), separate_docs_by_group=True,
            docnamePrefix='test',
            verbose=False, monitorPrefix='annotation docs')
        assert len(annotation_docs) == 3



import numpy


def exclude_nonfinite(m):
    """Exclude columns from m that have infinities or nans.  In the
    context of color deconvolution, these occur in conversion from RGB
    to SDA when the source has 0 in a channel.
    """
    return m[:, numpy.isfinite(m).all(axis=0)]



import numpy as np


def hessian(im_input, sigma):
    """
    Calculates hessian of image I convolved with a gaussian kernel with
    covariance C = [Sigma^2 0; 0 Sigma^2].

    Parameters
    ----------
    im_input : array_like
        M x N grayscale image.
    sigma : double
        standard deviation of gaussian kernel.

    Returns
    -------
    im_hess : array_like
        M x N x 4 hessian matrix - im_hess[:,:,0] = dxx,
        im_hess[:,:,1] = im_hess[:,:,2] = dxy, im_hess[:,:,3] = dyy.

    """
    from scipy.ndimage import convolve

    # generate kernel domain
    h, k = round(3 * sigma), round(3 * sigma + 1)
    x, y = np.mgrid[-h:k, -h:k]

    # generate kernels
    gxx = 1. / (2 * np.pi * sigma ** 4) * ((x / sigma) ** 2 - 1) * \
        np.exp(-(x**2 + y**2) / (2 * sigma ** 2))
    gxy = 1. / (2 * np.pi * sigma ** 6) * np.multiply(x, y) * \
        np.exp(-(x**2 + y**2) / (2 * sigma ** 2))
    gyy = np.transpose(gxx)

    # convolve
    dxx = convolve(im_input, gxx, mode='constant')
    dxy = convolve(im_input, gxy, mode='constant')
    dyy = convolve(im_input, gyy, mode='constant')

    # format output
    im_hess = np.concatenate(
        (dxx[:, :, None], dxy[:, :, None], dxy[:, :, None], dyy[:, :, None]),
        axis=2,
    )
    return im_hess



"""
This package contains functions for post-processing labeled segmentation
masks produced by segmentation algorithms.
"""

# must be imported after CondenseLabel
from .area_open import area_open
# make functions available at the package level using shadow imports
# since we mostly have one function per file
from .compact import compact
from .condense import condense
from .delete import delete
from .delete_border import delete_border
from .delete_overlap import delete_overlap
from .dilate_xor import dilate_xor
from .perimeter import perimeter
from .remove_overlap_nuclei import remove_overlap_nuclei
from .shuffle import shuffle
from .split import split
from .trace_object_boundaries import trace_object_boundaries
from .width_open import width_open

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'area_open',
    'compact',
    'dilate_xor',
    'condense',
    'delete',
    'delete_border',
    'perimeter',
    'shuffle',
    'split',
    'trace_object_boundaries',
    'width_open',
    'delete_overlap',
    'remove_overlap_nuclei',
)



import numpy as np


def del2(im_input):
    """Discrete Laplacian with edge-value extrapolation.

    Calculates the discrete Laplacian of an input image. Edge values are
    calculated by using linear extrapolation of second differences. This is
    consistent with the way that Matlab calculates the discrete Laplacian.

    Parameters
    ----------
    im_input : array_like
        A floating-point intensity image.

    Returns
    -------
    im_lap : array_like
        The discrete Laplacian of im_lap.

    See Also
    --------
    histomicstk.segmentation.level_set.reg_edge

    """
    # initialize output
    im_lap = np.zeros(im_input.shape)

    # process rows first
    D = np.diff(im_input, axis=0)
    S = np.zeros(im_input.shape)
    S[1:-1, :] = (D[1:, :] - D[0:-1, :]) / 2
    if im_input.shape[0] > 3:
        S[0, :] = 2 * S[1, :] - S[2, :]
        S[-1, :] = 2 * S[-2, :] - S[-3, :]
    elif im_input.shape[0] == 3:
        S[0, :] = S[1, :]
        S[-1, :] = S[1, :]
    else:
        S[0, :] = 0
        S[-1, :] = 0
    im_lap += S

    # process columns
    D = np.diff(im_input, axis=1)
    S = np.zeros(im_input.shape)
    S[:, 1:-1] = (D[:, 1:] - D[:, 0:-1]) / 2
    if im_input.shape[1] > 3:
        S[:, 0] = 2 * S[:, 1] - S[:, 2]
        S[:, -1] = 2 * S[:, -2] - S[:, -3]
    elif im_input.shape[1] == 3:
        S[0, :] = S[:, 1]
        S[:, -1] = S[:, 1]
    else:
        S[:, 0] = 0
        S[:, -1] = 0
    im_lap += S

    return im_lap / 2



import os
import shutil
import sys
import tempfile

import pytest
from pandas import read_csv

from histomicstk.annotations_and_masks.review_gallery import (
    _plot_rapid_review_vis, create_review_galleries,
    get_all_rois_from_folder_v2)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
import tests.htk_test_utilities as utilities  # noqa
from tests.htk_test_utilities import getTestFilePath, girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.URL = None
        self.GTCodes_dict = None
        self.BASE_SAVEPATH = None
        self.SAVEPATHS = None
        self.combinedvis_savepath = None
        self.folderid = None
        self.post_folderid = None


cfg = Cfg()

# pytest runs tests in the order they appear in the module


@pytest.mark.usefixtures('girderClient')  # noqa
def test_prep(girderClient):  # noqa

    cfg.gc = girderClient
    cfg.URL = cfg.gc.urlBase.replace('api/v1/', '')

    # GT codes dict for parsing into label mask
    gtcodePath = getTestFilePath('sample_GTcodes.csv')
    cfg.GTCodes_dict = read_csv(gtcodePath)
    cfg.GTCodes_dict.index = cfg.GTCodes_dict.loc[:, 'group']
    cfg.GTCodes_dict = cfg.GTCodes_dict.to_dict(orient='index')

    # just a temp directory to save masks
    cfg.BASE_SAVEPATH = tempfile.mkdtemp()
    cfg.SAVEPATHS = {
        'contours': os.path.join(cfg.BASE_SAVEPATH, 'contours'),
        'rgb': os.path.join(cfg.BASE_SAVEPATH, 'rgbs'),
        'visualization': os.path.join(cfg.BASE_SAVEPATH, 'vis'),
    }
    for _, savepath in cfg.SAVEPATHS.items():
        os.mkdir(savepath)

    # where to save gallery
    cfg.combinedvis_savepath = os.path.join(cfg.BASE_SAVEPATH, 'combinedvis')
    os.mkdir(cfg.combinedvis_savepath)

    # get original item
    iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

    # create the folder to parse its contents into galleries
    folderinfo = cfg.gc.post(
        '/folder', data={
            'parentId': iteminfo['folderId'],
            'name': 'test-review',
        })
    cfg.folderid = folderinfo['_id']

    # create girder folder to post resultant galleries
    post_folderinfo = cfg.gc.post(
        '/folder', data={
            'parentId': iteminfo['folderId'],
            'name': 'test-review-post',
        })
    cfg.post_folderid = post_folderinfo['_id']

    # copy the item multiple times to create dummy database
    for i in range(2):
        cfg.gc.post(
            '/item/%s/copy' % iteminfo['_id'], data={
                'name': 'testSlide%dForRevGal' % i,
                'copyAnnotations': True,
                'folderId': cfg.folderid,
            })


class TestReviewGallery:
    """Test methods for getting review gallery."""

    def test_get_all_rois_from_folder_v2(self):
        """Test get_all_rois_from_folder_v2()."""
        # params for getting all rois for slide
        get_all_rois_kwargs = {
            'GTCodes_dict': cfg.GTCodes_dict,
            'save_directories': cfg.SAVEPATHS,
            'annotations_to_contours_kwargs': {
                'MPP': 5.0,
                'linewidth': 0.2,
                'get_rgb': True,
                'get_visualization': True,
            },
            'verbose': False,
            'get_mask': False,
            # we use this callback so that we have results compatible
            # of being used as input for create_review_galleries()
            'callback': _plot_rapid_review_vis,
            'callback_kwargs': {
                'combinedvis_savepath': cfg.combinedvis_savepath,
                'zoomout': 1.5,

            },
        }

        # Get al rois to prep for gallery
        get_all_rois_from_folder_v2(
            gc=cfg.gc, folderid=cfg.folderid,
            get_all_rois_kwargs=get_all_rois_kwargs, monitor='test')

        # a couple of checks
        all_fovs = os.listdir(cfg.combinedvis_savepath)

        fovdict = {}
        for fov in all_fovs:
            slide = fov.split('_')[0]
            if slide not in fovdict.keys():
                fovdict[slide] = []
            fovdict[slide].append(fov.split(slide)[1].split('.png')[0])

        assert set(fovdict.keys()) == {
            'testSlide0ForRevGal', 'testSlide1ForRevGal'}

        assert {
            j.split('_left-')[1] for j in fovdict['testSlide0ForRevGal']
        } == {
            '%d_top-%d_bottom-%d_right-%d'
            % (l, t, b, r) for (l, t, b, r) in [
                (57584, 35788, 37425, 59421),
                (58463, 38203, 39760, 60379),
                (59181, 33473, 38043, 63712),
            ]
        }

    def test_create_review_galleries(self):
        """Test create_review_galleries()."""
        create_review_galleries_kwargs = {
            'tilepath_base': cfg.combinedvis_savepath,
            'upload_results': True,
            'gc': cfg.gc,
            'url': cfg.URL,
            'gallery_folderid': cfg.post_folderid,
            'gallery_savepath': None,
            'padding': 25,
            'tiles_per_row': 2,
            'tiles_per_column': 5,
            'nameprefix': 'test',
        }

        # create (+/- post) review gallery
        resps = create_review_galleries(**create_review_galleries_kwargs)

        assert len(resps) == 1
        assert {j['name'] for j in resps} == {'test_gallery-1'}

        # Now you can go to the histomicstk folder POST_FOLDERID
        # to check that the visualizations and gallery make sense!

        # cleanup
        shutil.rmtree(cfg.BASE_SAVEPATH)



"""
Created on Sat Oct 19 18:09:17 2019.

@author: mtageld
"""
import numpy as np

from histomicstk.preprocessing.color_conversion import rgb_to_sda, sda_to_rgb
from histomicstk.preprocessing.color_deconvolution import \
    color_deconvolution_routine
from histomicstk.utils import convert_image_to_matrix, convert_matrix_to_image


def perturb_stain_concentration(
        StainsFloat, W, I_0=None, mask_out=None, sigma1=0.9, sigma2=0.9):
    """Perturb stain concentrations in SDA space and return augmented image.

    This is an implementation of the method described in Tellez et
    al, 2018 (see below). The SDA matrix is perturbed by multiplying each
    channel independently by a value chosen from a random uniform distribution
    in the range [1 - sigma1, 1 + sigma1], then add a value chosen from another
    random uniform distribution in the range [-sigma2, sigma2].

    Parameters
    ----------
    StainsFloat : array_like
        An intensity image (m, n, 3) of deconvolved stains that is unbounded,
        suitable for reconstructing color images of deconvolved stains
        with color_convolution.

    W : array_like
        A 3x3 complemented stain matrix.

    I_0 : float or array_like, optional
        A float a 3-vector containing background RGB intensities.
        If unspecified, use the old OD conversion.

    mask_out : array_like, default is None
        if not None, should be (m x n) boolean numpy array.
        This parameter ensures exclusion of non-masked areas from perturbing.
        This is relevant because elements like blood, sharpie marker,
        white space, etc cannot be simply modeled as a mix of two stains.

    sigma1 : float
        parameter, see beginning of this docstring.

    sigma2 : float
        parameter, see beginning of this docstring.

    Returns
    -------
    array_like
        Color augmented RGB image (m x n x 3)

    References
    ----------
    .. [#] Tellez, David, Maschenka Balkenhol, Irene Otte-Höller,
           Rob van de Loo, Rob Vogels, Peter Bult, Carla Wauters et al.
           "Whole-slide mitosis detection in H&E breast histology using PHH3
           as a reference to train distilled stain-invariant convolutional
           networks." IEEE transactions on medical imaging 37, no. 9
           (2018): 2126-2136.
    .. [#] Tellez, David, Geert Litjens, Peter Bandi, Wouter Bulten,
           John-Melle Bokhorst, Francesco Ciompi, and Jeroen van der Laak.
           "Quantifying the effects of data augmentation and stain color
           normalization in convolutional neural networks for computational
           pathology." arXiv preprint arXiv:1902.06543 (2019).
    .. [#] Implementation inspired by Peter Byfield StainTools repository. See
           https://github.com/Peter554/StainTools/blob/master/LICENSE.txt
           for copyright license (MIT license).

    """
    # augment everything, otherwise only augment specific pixels
    if mask_out is None:
        keep_mask = np.zeros(StainsFloat.shape[:2]) == 0
    else:
        keep_mask = np.equal(mask_out, False)
    keep_mask = np.tile(keep_mask[..., None], (1, 1, 3))
    keep_mask = convert_image_to_matrix(keep_mask)

    # transform 3D input stain image to 2D stain matrix format
    m = convert_image_to_matrix(StainsFloat)

    # transform input stains to optical density values
    sda_fwd = rgb_to_sda(
        m, 255 if I_0 is not None else None, allow_negatives=True)

    # perturb concentrations in SDA space
    augmented_sda = sda_fwd.copy()
    for i in range(3):
        alpha = np.random.uniform(1 - sigma1, 1 + sigma1)
        beta = np.random.uniform(-sigma2, sigma2)
        augmented_sda[i, keep_mask[i, :]] *= alpha
        augmented_sda[i, keep_mask[i, :]] += beta

    # convolve with stains column vectors and convert to RGB
    sda_conv = np.dot(W, augmented_sda)
    sda_inv = sda_to_rgb(sda_conv, I_0)

    # reshape output, transform type
    augmented_rgb = (
        convert_matrix_to_image(sda_inv, StainsFloat.shape)
        .clip(0, 255).astype(np.uint8))

    return augmented_rgb


def rgb_perturb_stain_concentration(
        im_rgb, stain_unmixing_routine_params=None, **kwargs):
    """Apply wrapper that calls perturb_stain_concentration() on RGB.

    Parameters
    ----------
    im_rgb : array_like
        An RGB image (m x n x 3) to color normalize

    stain_unmixing_routine_params : dict
        kwargs to pass as-is to the color_deconvolution_routine().

    kwargs : k,v pairs
        Passed as-is to perturb_stain_concentration()

    Returns
    -------
    array_like
        Color augmented RGB image (m x n x 3)

    """
    stain_unmixing_routine_params = {
        'stains': ['hematoxylin', 'eosin'],
        'stain_unmixing_method': 'macenko_pca',
    } if stain_unmixing_routine_params is None else stain_unmixing_routine_params

    _, StainsFloat, W_source = color_deconvolution_routine(
        im_rgb, W_source=None, **stain_unmixing_routine_params)

    return perturb_stain_concentration(StainsFloat, W_source, **kwargs)



def sda_to_rgb(im_sda, I_0):
    """Transform input SDA image or matrix `im_sda` into RGB space.  This
    is the inverse of `rgb_to_sda` with respect to the first parameter

    Parameters
    ----------
    im_sda : array_like
        Image (MxNx3) or matrix (3xN) of pixels

    I_0 : float or array_like
        Background intensity, either per-channel or for all channels

    Note
    ----
    For compatibility purposes, passing I_0=None invokes the behavior of
    od_to_rgb.

    See Also
    --------
    histomicstk.preprocessing.color_conversion.rgb_to_sda,
    histomicstk.preprocessing.color_conversion.od_to_rgb,
    histomicstk.preprocessing.color_deconvolution.color_deconvolution,
    histomicstk.preprocessing.color_deconvolution.color_convolution

    """
    is_matrix = im_sda.ndim == 2
    if is_matrix:
        im_sda = im_sda.T

    od = I_0 is None
    if od:  # od_to_rgb compatibility
        I_0 = 256

    im_rgb = I_0 ** (1 - im_sda / 255.)
    return (im_rgb.T if is_matrix else im_rgb) - od



import shapely


def create_polygon(coordinates):
    """
    Create a shapely Polygon from a list of points.

    Args:
    ----
        points (list): A list containing tuples representing the points of the polygon.

    Returns:
    -------
        shapely.geometry.Polygon: The polygon created from the points.

    """
    return shapely.geometry.Polygon(coordinates).buffer(0)


def convert_polygons_tobbox(nuclei_list):
    """
    Convert nuclei segmentation data from polygon format to bounding box format.

    Parameters
    ----------
    nuclei_list (list): A list of dictionaries, where each dictionary represents a nuclei object.

    Returns
    -------
    list: The modified 'nuclei_list', where each nuclei object has been transformed
    to a bounding box representation.

    """
    for nuclei in nuclei_list:
        # Set 'type' property of the nuclei to 'rectangle'
        nuclei['type'] = 'rectangle'
        nuclei['rotation'] = 0

        # Calculate the minimum and maximum x and y coordinates to obtain the bounding box.
        minx = min(p[0] for p in nuclei['points'])
        maxx = max(p[0] for p in nuclei['points'])
        miny = min(p[1] for p in nuclei['points'])
        maxy = max(p[1] for p in nuclei['points'])

        # Calculate the center of the bounding box, width, height
        nuclei['center'] = [(minx + maxx) / 2, (miny + maxy) / 2, 0]
        nuclei['width'] = maxx - minx
        nuclei['height'] = maxy - miny

        # Remove the 'closed' and 'points'
        nuclei.pop('closed')
        nuclei.pop('points')

    return nuclei_list


def remove_overlap_nuclei(nuclei_list, nuclei_format, return_selected_nuclei=False):
    """
    Remove overlapping nuclei using spatial indexing and parallel processing.

    This function removes overlapping nuclei polygons from input using an STRtree index.

    Args:
    ----
        nuclei_list (list): List of dictionaries with 'points' for polygons.
        nuclei_format (str, optional): Output format ('polygon' or 'bbox').
        return_selected (bool, optional): Return indices of selected nuclei (default: False).

    Returns:
    -------
        output_list (list): New list with overlapping nuclei removed.
        selected_nuclei (list, optional): Indices of selected nuclei.

    """
    polygons = [create_polygon(nuclei['points']) for nuclei in nuclei_list]

    # Build the STRtree from all the polygons
    rt = shapely.strtree.STRtree(polygons)

    # Find and remove any overlapping polygons
    output = [(nuclei, index) for index, nuclei in enumerate(nuclei_list) if not any(
        ix for ix in rt.query(polygons[index])
        if ix < index and polygons[index].intersects(polygons[ix]))]

    output_list, selected_nuclei = (
        zip(*output) if return_selected_nuclei else ([nuclei for nuclei, _ in output], []))

    # if nuclei_format is bbox - convert polygons to bbox
    if nuclei_format == 'bbox':
        output_list = convert_polygons_tobbox(output_list)

    if return_selected_nuclei:
        return output_list, selected_nuclei

    return output_list



import numpy as np


def simple_mask(im_rgb, bandwidth=2, bgnd_std=2.5, tissue_std=30,
                min_peak_width=10, max_peak_width=25,
                fraction=0.10, min_tissue_prob=0.05):
    """Performs segmentation of the foreground (tissue)
    Uses a simple two-component Gaussian mixture model to mask tissue areas
    from background in brightfield H&E images. Kernel-density estimation is
    used to create a smoothed image histogram, and then this histogram is
    analyzed to identify modes corresponding to tissue and background. The
    mode peaks are then analyzed to estimate their width, and a constrained
    optimization is performed to fit gaussians directly to the histogram
    (instead of using expectation-maximization directly on the data which
    is more prone to local minima effects). A maximum-likelihood threshold
    is then derived and used to mask the tissue area in a binarized image.

    Parameters
    ----------
    im_rgb : array_like
        An RGB image of type unsigned char.
    bandwidth : double, optional
        Bandwidth for kernel density estimation - used for smoothing the
        grayscale histogram. Default value = 2.
    bgnd_std : double, optional
        Standard deviation of background gaussian to be used if
        estimation fails. Default value = 2.5.
    tissue_std: double, optional
        Standard deviation of tissue gaussian to be used if estimation fails.
        Default value = 30.
    min_peak_width: double, optional
        Minimum peak width for finding peaks in KDE histogram. Used to
        initialize curve fitting process. Default value = 10.
    max_peak_width: double, optional
        Maximum peak width for finding peaks in KDE histogram. Used to
        initialize curve fitting process. Default value = 25.
    fraction: double, optional
        Fraction of pixels to sample for building foreground/background
        model. Default value = 0.10.
    min_tissue_prob : double, optional
        Minimum probability to qualify as tissue pixel. Default value = 0.05.

    Returns
    -------
    im_mask : array_like
        A binarized version of `I` where foreground (tissue) has value '1'.

    See Also
    --------
    histomicstk.utils.sample_pixels

    """
    from scipy import signal
    from scipy.optimize import fmin_slsqp
    from scipy.stats import norm
    from skimage import color
    from sklearn.neighbors import KernelDensity

    # convert image to grayscale, flatten and sample
    im_rgb = 255 * color.rgb2gray(im_rgb)
    im_rgb = im_rgb.astype(np.uint8)
    num_samples = int(fraction * im_rgb.size)
    sI = np.random.choice(im_rgb.flatten(), num_samples)[:, np.newaxis]

    # kernel-density smoothed histogram
    KDE = KernelDensity(kernel='gaussian', bandwidth=bandwidth).fit(sI)
    xHist = np.linspace(0, 255, 256)[:, np.newaxis]
    yHist = np.exp(KDE.score_samples(xHist))[:, np.newaxis]
    yHist = yHist / sum(yHist)

    # flip smoothed y-histogram so that background mode is on the left side
    yHist = np.flipud(yHist)

    # identify initial mean parameters for gaussian mixture distribution
    # take highest peak among remaining peaks as background
    Peaks = signal.find_peaks_cwt(yHist.flatten(),
                                  np.arange(min_peak_width, max_peak_width))
    BGPeak = Peaks[0]
    if len(Peaks) > 1:
        TissuePeak = Peaks[yHist[Peaks[1:]].argmax() + 1]
    else:  # no peak found - take initial guess at 2/3 distance from origin
        TissuePeak = xHist[int(np.round(0.66 * xHist.size))].item()

    # analyze background peak to estimate variance parameter via FWHM
    BGScale = estimate_variance(xHist, yHist, BGPeak)
    if BGScale == -1:
        BGScale = bgnd_std

    # analyze tissue peak to estimate variance parameter via FWHM
    TissueScale = estimate_variance(xHist, yHist, TissuePeak)
    if TissueScale == -1:
        TissueScale = tissue_std

    # solve for mixing parameter
    Mix = yHist[BGPeak] * (BGScale * (2 * np.pi)**0.5)
    try:
        if len(Mix) == 1:
            Mix = Mix[0]
    except Exception:
        pass

    # flatten kernel-smoothed histogram and corresponding x values for
    # optimization
    xHist = xHist.flatten()
    yHist = yHist.flatten()

    # define gaussian mixture model
    def gaussian_mixture(x, mu1, mu2, sigma1, sigma2, p):
        rv1 = norm(loc=mu1, scale=sigma1)
        rv2 = norm(loc=mu2, scale=sigma2)
        return p * rv1.pdf(x) + (1 - p) * rv2.pdf(x)

    # define gaussian mixture model residuals
    def gaussian_residuals(Parameters, y, x):
        mu1, mu2, sigma1, sigma2, p = Parameters
        yhat = gaussian_mixture(x, mu1, mu2, sigma1, sigma2, p)
        return sum((y - yhat) ** 2)

    # fit Gaussian mixture model and unpack results
    Parameters = fmin_slsqp(gaussian_residuals,
                            [BGPeak, TissuePeak, BGScale, TissueScale, Mix],
                            args=(yHist, xHist),
                            bounds=[(0, 255), (0, 255),
                                    (np.spacing(1), 10),
                                    (np.spacing(1), 50), (0, 1)],
                            iprint=0)

    muBackground = Parameters[0]
    muTissue = Parameters[1]
    sigmaBackground = Parameters[2]
    sigmaTissue = Parameters[3]
    p = Parameters[4]

    # create mask based on Gaussian mixture model
    Background = norm(loc=muBackground, scale=sigmaBackground)
    Tissue = norm(loc=muTissue, scale=sigmaTissue)
    pBackground = p * Background.pdf(xHist)
    pTissue = (1 - p) * Tissue.pdf(xHist)

    # identify maximum likelihood threshold
    Difference = pTissue - pBackground
    Candidates = np.nonzero(Difference >= 0)[0]
    Filtered = np.nonzero(xHist[Candidates] > muBackground)
    ML = xHist[Candidates[Filtered[0]][0]]

    # identify limits for tissue model (MinProb, 1-MinProb)
    Endpoints = np.asarray(Tissue.interval(1 - min_tissue_prob / 2))

    # invert threshold and tissue mean
    ML = 255 - ML
    muTissue = 255 - muTissue
    Endpoints = np.sort(255 - Endpoints)

    # generate mask
    im_mask = (im_rgb <= ML) & (im_rgb >= Endpoints[0]) & \
              (im_rgb <= Endpoints[1])
    im_mask = im_mask.astype(np.uint8)

    return im_mask


def estimate_variance(x, y, peak):
    """Estimates variance of a peak in a histogram using the FWHM of an
    approximate normal distribution.
    Starting from a user-supplied peak and histogram, this method traces down
    each side of the peak to estimate the full-width-half-maximum (FWHM) and
    variance of the peak. If tracing fails on either side, the FWHM is
    estimated as twice the HWHM.

    Parameters
    ----------
    x : array_like
        vector of x-histogram locations.
    y : array_like
        vector of y-histogram locations.
    peak : double
        index of peak in y to estimate variance of

    Returns
    -------
    scale : double
        Standard deviation of normal distribution approximating peak. Value is
        -1 if fitting process fails.

    See Also
    --------
    SimpleMask

    """
    # analyze peak to estimate variance parameter via FWHM
    peak = int(peak)
    Left = peak
    while y[Left] > y[peak] / 2 and Left >= 0:
        Left -= 1
        if Left == -1:
            break
    Right = peak
    while y[Right] > y[peak] / 2 and Right < y.size:
        Right += 1
        if Right == y.size:
            break
    if Left != -1 and Right != y.size:
        LeftSlope = y[Left + 1] - y[Left] / (x[Left + 1] - x[Left])
        Left = (y[peak] / 2 - y[Left]) / LeftSlope + x[Left]
        RightSlope = y[Right] - y[Right - 1] / (x[Right] - x[Right - 1])
        Right = (y[peak] / 2 - y[Right]) / RightSlope + x[Right]
        scale = (Right - Left) / 2.355
    if Left == -1:
        if Right == y.size:
            scale = -1
        else:
            RightSlope = y[Right] - y[Right - 1] / (x[Right] - x[Right - 1])
            Right = (y[peak] / 2 - y[Right]) / RightSlope + x[Right]
            scale = 2 * (Right - x[peak]) / 2.355
    if Right == y.size:
        if Left == -1:
            scale = -1
        else:
            LeftSlope = y[Left + 1] - y[Left] / (x[Left + 1] - x[Left])
            Left = (y[peak] / 2 - y[Left]) / LeftSlope + x[Left]
            scale = 2 * (x[peak] - Left) / 2.355
    try:
        if len(scale) == 1:
            scale = scale[0]
    except Exception:
        pass
    return scale



from .rgb_to_sda import rgb_to_sda


def rgb_to_od(im_rgb):
    """Transforms input RGB image `im_rgb` into optical density space
    for color deconvolution.

    Parameters
    ----------
    im_rgb : array_like
        A floating-point RGB image with intensity ranges of [0, 255].

    Returns
    -------
    im_od : array_like
        A floating-point image of corresponding optical density values.


    See Also
    --------
    histomicstk.preprocessing.color_conversion.od_to_rgb,
    histomicstk.preprocessing.color_conversion.rgb_to_sda,
    histomicstk.preprocessing.color_deconvolution.color_deconvolution,
    histomicstk.preprocessing.color_deconvolution.color_convolution

    """
    return rgb_to_sda(im_rgb, None)  # Compatibility mode



"""
Created on Wed Aug 21 16:25:06 2019.

@author: tageldim

"""
import os

import numpy as np
from imageio import imread
from PIL import Image
# import cv2
from shapely.geometry.polygon import Polygon
from shapely.ops import unary_union

from histomicstk.annotations_and_masks.masks_to_annotations_handler import (
    _discard_nonenclosed_background_group, _parse_annot_coords,
    get_contours_from_mask)
from histomicstk.utils.general_utils import Base_HTK_Class


class Polygon_merger(Base_HTK_Class):
    """Methods to merge polygons in tiled masks."""

    def __init__(self, maskpaths, GTCodes_df, **kwargs):
        """Init Polygon_merger object.

        Arguments:
        ---------
        maskpaths : list
            list of strings representing pathos to masks
        GTCodes_df : pandas DataFrame
            the ground truth codes and information dataframe.
            This is a dataframe that MUST BE indexed by annotation group name
            and has the following columns.

            group: str
                group name of annotation, eg. mostly_tumor.
            GT_code: int
                desired ground truth code (in the mask). Pixels of this value
                belong to corresponding group (class).
            color: str
                rgb format. eg. rgb(255,0,0).
        imreader : function
            function that takes a maskpath and returns mask. Default is
            imageio.imread
        imreader_kws : dict
            kwargs for mask reader
        merge_thresh : int
            how close do the polygons need to be (in pixels) to be merged
        contkwargs : dict
            dictionary of kwargs to pass to get_contours_from_mask()
        discard_nonenclosed_background : bool
            If a background group contour is NOT fully enclosed, discard it.
            This is a purely aesthetic method, makes sure that the background
            contours (eg stroma) are discarded by default to avoid cluttering
            the field when posted to DSA for viewing online. The only exception
            is if they are enclosed within something else (eg tumor), in which
            case they are kept since they represent holes. This is related to
            https://github.com/DigitalSlideArchive/HistomicsTK/issues/675
            WARNING - This is a bit slower since the contours will have to be
            converted to shapely polygons. It is not noticeable for hundreds of
            contours, but you will notice the speed difference if you are
            parsing thousands of contours. Default, for this reason, is False.
        background_group : str
            name of bckgrd group in the GT_codes dataframe (eg mostly_stroma)
        roi_group : str
            name of roi group in the GT_Codes dataframe (eg roi)
        verbose : int
            0 - Do not print to screen
            1 - Print only key messages
            2 - Print everything to screen
        monitorPrefix : str
            text to prepend to printed statements

        """
        self.maskpaths = maskpaths
        self.GTCodes_df = GTCodes_df

        # must have an roi group color for merged polygon bbox
        if 'roi' not in self.GTCodes_df.index:
            self.GTCodes_df.loc['roi', 'color'] = 'rgb(0,0,0)'

        # see: https://stackoverflow.com/questions/8187082/how-can-you-set-...
        # class-attributes-from-variable-arguments-kwargs-in-python

        contkwargs = {
            'GTCodes_df': GTCodes_df,
            'get_roi_contour': False,  # important
            'discard_nonenclosed_background': False,  # important
            'MIN_SIZE': 2,
            'MAX_SIZE': None,
        }
        if 'contkwargs' in kwargs.keys():
            contkwargs.update(kwargs['contkwargs'])
        kwargs['contkwargs'] = contkwargs

        default_attr = {
            'verbose': 1,
            'monitorPrefix': '',
            'imreader': imread,
            'imreader_kws': {},
            'merge_thresh': 3,
            'discard_nonenclosed_background': False,
            'background_group': 'mostly_stroma',
            'roi_group': 'roi',
        }
        default_attr.update(kwargs)
        super().__init__(default_attr=default_attr)

        # some sanity checks
        assert not (
            self.contkwargs['get_roi_contour'])
        assert not (
            self.contkwargs['discard_nonenclosed_background'])
        self.contkwargs['verbose'] = self.verbose > 1

    def set_contours_from_all_masks(self, monitorPrefix=''):
        """Get contours_df from all masks.

        This is a wrapper around get_contours_from_mask(), with the added
        functionality of separating out contorus at roi edge from those that
        are not.

        Sets:
        - self.ordinary_contours: dict: indexed by maskname, each entry
        is a contours dataframe
        - self.edge_contours: dict: indexed by maskname, each entry is
        a contours dataframe
        - self.merged_contours: pandas DataFrame: single dataframe to
        save all merged contours

        """
        from pandas import DataFrame

        ordinary_contours = {}
        edge_contours = {}

        to_remove = []

        for midx, maskpath in enumerate(self.maskpaths):

            # read mask
            MASK = self.imreader(maskpath, **self.imreader_kws)

            # mask is empty!
            if MASK.sum() < 2:
                to_remove.append(maskpath)
                continue

            # extract contours
            contours_df = get_contours_from_mask(
                MASK=MASK,
                monitorPrefix='%s: mask %d of %d' % (
                    monitorPrefix, midx, len(self.maskpaths)),
                **self.contkwargs)

            # no contours!
            if contours_df.shape[0] < 1:
                to_remove.append(maskpath)
                continue

            # separate edge from non-edge contours
            edgeids = []
            for edge in ['top', 'left', 'bottom', 'right']:
                edgeids.extend(list(contours_df.loc[contours_df.loc[
                    :, 'touches_edge-%s' % edge] == 1, :].index))
            edgeids = list(set(edgeids))
            roiname = os.path.split(maskpath)[1]
            edge_contours[roiname] = contours_df.loc[edgeids, :].copy()
            ordinary_contours[roiname] = contours_df.drop(edgeids, axis=0)

        self.maskpaths = [j for j in self.maskpaths if j not in to_remove]

        self.ordinary_contours = ordinary_contours
        self.edge_contours = edge_contours
        # init dataframe to save merged contours
        colnames = edge_contours[list(edge_contours.keys())[0]].columns
        self.merged_contours = DataFrame(columns=colnames)

    def _get_mask_offsets_from_masknames(self):
        """Get dictionary of mask offsets (top and left) (Internal).

        The pattern _left-123_ and _top-123_ is assumed to
        encode the x and y offset of the mask at base magnification.

        Returns
        -------
        - dict: indexed by maskname, each entry is a dict with keys
        top and left.

        """
        roi_offsets = {}
        for maskpath in self.maskpaths:
            maskname = os.path.split(maskpath)[1]
            roi_offsets[maskname] = {
                'left': int(maskname.split('_left-')[1].split('_')[0]),
                'top': int(maskname.split('_top-')[1].split('_')[0]),
            }
        return roi_offsets

    def set_roi_bboxes(self, roi_offsets=None):
        """Get dictionary of roi bounding boxes.

        Arguments:
        ---------
        - roi_offsets: dict (default, None): dict indexed by maskname,
        each entry is a dict with keys top and left each is an integer.
        If None, then the x and y offset is inferred from mask name.

        Sets:
        - self.roiinfos: dict: dict indexed by maskname, each entry is a
        dict with keys top, left, bottom, right, all of which are integers.

        """
        if roi_offsets is not None:
            roiinfos = roi_offsets.copy()
        else:
            # get offset for all rois. This result is a dict that is indexed
            # by maskname, each entry is a dict with keys 'top' and 'left'.
            roiinfos = self._get_mask_offsets_from_masknames()

        for maskpath in self.maskpaths:
            # Note: the following method does NOT actually load the mask
            # but just uses pillow to get its metadata. See:
            # https://stackoverflow.com/questions/15800704/ ...
            # ... get-image-size-without-loading-image-into-memory
            with Image.open(maskpath, mode='r') as mask_obj:
                width, height = mask_obj.size[:2]
                maskname = os.path.split(maskpath)[1]
                roiinfos[maskname]['right'] = roiinfos[
                    maskname]['left'] + width
                roiinfos[maskname]['bottom'] = roiinfos[
                    maskname]['top'] + height

        self.roiinfos = roiinfos

    def _get_roi_pairs(self):
        """Get unique roi pairs (Internal)."""
        ut = np.triu_indices(len(self.roinames), k=1)
        roi_pairs = []
        for pairidx in range(len(ut[0])):
            roi_pairs.append((ut[0][pairidx], ut[1][pairidx]))
        return roi_pairs

    def set_shared_roi_edges(self):
        """Get shared edges between rois in same slide (Internal)."""
        from pandas import DataFrame

        self.roinames = list(self.roiinfos.keys())
        edgepairs = [
            ('left', 'right'), ('right', 'left'),
            ('top', 'bottom'), ('bottom', 'top'),
        ]
        roi_pairs = self._get_roi_pairs()

        # init shared edges
        shared_edges = DataFrame(columns=[
            'roi1-name', 'roi1-edge', 'roi2-name', 'roi2-edge'])

        for roi_pair in roi_pairs:
            roi1name = self.roinames[roi_pair[0]]
            roi2name = self.roinames[roi_pair[1]]
            idx = shared_edges.shape[0]
            for edgepair in edgepairs:
                # check if they share bounds for one edge
                if np.abs(
                    self.roiinfos[roi1name][edgepair[0]]
                        - self.roiinfos[roi2name][edgepair[1]]) < 2:
                    # ensure they overlap in location along other axis
                    if 'left' in edgepair:
                        start, end = ('top', 'bottom')
                    else:
                        start, end = ('left', 'right')
                    realStart = np.min(
                        (self.roiinfos[roi1name][start],
                         self.roiinfos[roi2name][start]))
                    realEnd = np.max(
                        (self.roiinfos[roi1name][end],
                         self.roiinfos[roi2name][end]))
                    length = realEnd - realStart
                    nonoverlap_length = (
                        self.roiinfos[roi1name][end]
                        - self.roiinfos[roi1name][start]) + (
                        self.roiinfos[roi2name][end]
                        - self.roiinfos[roi2name][start])
                    if length < nonoverlap_length:
                        shared_edges.loc[idx, 'roi1-name'] = roi1name
                        shared_edges.loc[idx, 'roi1-edge'] = edgepair[0]
                        shared_edges.loc[idx, 'roi2-name'] = roi2name
                        shared_edges.loc[idx, 'roi2-edge'] = edgepair[1]

        self.shared_edges = shared_edges

    def _get_merge_pairs(self, edgepair, group, monitorPrefix=''):
        """Get nest dfs and indices of which ones to merge (Internal)."""
        from pandas import DataFrame

        def _get_nests_slice(ridx=1):
            Nests = self.edge_contours[edgepair['roi%d-name' % ridx]]
            edgevar = 'touches_edge-%s' % (edgepair['roi%d-edge' % ridx])
            Nests = Nests.loc[Nests.loc[:, edgevar] == 1, :]
            Nests = Nests.loc[Nests.loc[:, 'group'] == group, :]
            return Nests

        # Nests of the same label. The nest IDs are using the index of the
        # roi dataframe from the edge_nests dictionary
        Nests1 = _get_nests_slice(1)
        Nests2 = _get_nests_slice(2)

        # to avoid redoing things, keep all polygons in a list
        polygons1 = []
        nno1 = 0
        nno1Max = Nests1.shape[0]
        for nid1, nest1 in Nests1.iterrows():
            nno1 += 1
            self._print2('%s: edge1-nest %d of %d' % (
                monitorPrefix, nno1, nno1Max))
            try:
                coords = np.array(_parse_annot_coords(nest1))
                coords[:, 0] = coords[:, 0] + self.roiinfos[
                    edgepair['roi1-name']]['left']
                coords[:, 1] = coords[:, 1] + self.roiinfos[
                    edgepair['roi1-name']]['top']
                polygons1.append((nid1, Polygon(coords)))
            except Exception as e:
                self._print2(
                    '%s: edge1-nest %d of %d: Shapely Error (below)' % (
                        monitorPrefix, nno1, nno1Max))
                self._print2(e)

        # go through the "other" polygons to get merge list
        to_merge = DataFrame(columns=[
            'nest1-roiname', 'nest1-nid', 'nest2-roiname', 'nest2-nid'])
        nno2 = 0
        nno2Max = Nests2.shape[0]
        for nid2, nest2 in Nests2.iterrows():
            nno2 += 1
            try:
                coords = np.array(_parse_annot_coords(nest2))
                coords[:, 0] = coords[:, 0] + self.roiinfos[
                    edgepair['roi2-name']]['left']
                coords[:, 1] = coords[:, 1] + self.roiinfos[
                    edgepair['roi2-name']]['top']
                polygon2 = Polygon(coords)
            except Exception as e:
                self._print2(
                    '%s: edge2-nest %d of %d: Shapely Error (below)' % (
                        monitorPrefix, nno2, nno2Max))
                self._print2(e)
                continue

            nno1Max = len(polygons1) - 1
            for nno1, poly1 in enumerate(polygons1):
                self._print2(
                    '%s: edge2-nest %d of %d: vs. edge1-nest %d of %d' % (
                        monitorPrefix, nno2, nno2Max, nno1 + 1, nno1Max + 1))
                nid1, polygon1 = poly1
                if polygon1.distance(polygon2) < self.merge_thresh:
                    idx = to_merge.shape[0]
                    to_merge.loc[idx, 'nest1-roiname'] = edgepair['roi1-name']
                    to_merge.loc[idx, 'nest1-nid'] = nid1
                    to_merge.loc[idx, 'nest2-roiname'] = edgepair['roi2-name']
                    to_merge.loc[idx, 'nest2-nid'] = nid2

        return to_merge

    def _get_merge_df(self, group, monitorPrefix=''):
        """Get merge dataframe (pairs along shared edges) (Internal)."""
        from pandas import DataFrame, concat

        merge_df = DataFrame()
        for rpno, edgepair in self.shared_edges.iterrows():
            edgepair = dict(edgepair)
            to_merge = self._get_merge_pairs(
                edgepair=edgepair, group=group,
                monitorPrefix='%s: pair %d of %d' % (
                    monitorPrefix, rpno + 1, self.shared_edges.shape[0]))
            merge_df = concat((merge_df, to_merge), axis=0, ignore_index=True)
        return merge_df

    def _get_merge_clusters_from_df(self, merge_df, monitorPrefix=''):
        """Assign each nest to one cluster (Internal).

        That is, such that all nests that are connected to each other are
        in the same cluster. Uses hierarchical-like clustering to do this.

        """
        checksum_ref = []

        # initi merge groups with clusters of size 2
        merge_groups = {'level-0': []}
        for _, merge in merge_df.iterrows():
            nname1 = '%s_nid-%d' % (merge['nest1-roiname'], merge['nest1-nid'])
            nname2 = '%s_nid-%d' % (merge['nest2-roiname'], merge['nest2-nid'])
            merge_groups['level-0'].append([nname1, nname2])
            checksum_ref.extend([nname1, nname2])
        checksum_ref = len(list(set(checksum_ref)))

        # go through levels
        level = 0
        keep_going = True
        while keep_going:
            self._print2('%s: level %d' % (monitorPrefix, level))
            merge_groups['level-%d' % (level + 1)] = []
            reference = merge_groups['level-%d' % (level)][:]
            # compare each cluster to the others in current level
            for gid, mgroup in enumerate(merge_groups['level-%d' % (level)]):
                for gid2, mgroup2 in enumerate(reference):
                    mgroupSet = set(mgroup)
                    mgroup2Set = set(mgroup2)
                    # if there's anything in common between the two clusters
                    # (assuming they're not the same cluster of course)
                    # then merge them and move them to the next level in the
                    # hierarchy and remove them both from this level
                    if (gid != gid2) and (
                            len(mgroupSet.intersection(mgroup2Set)) > 0):
                        merge_groups['level-%d' % (level + 1)].append(
                            list(mgroupSet.union(set(mgroup2Set))))
                        reference[gid] = []
                        reference[gid2] = []
            # cleanup
            merge_groups['level-%d' % (level)] = [
                j for j in reference if len(j) > 0]
            if len(merge_groups['level-%d' % (level + 1)]) < 1:
                del merge_groups['level-%d' % (level + 1)]
                keep_going = False
            level += 1

        # now just concatenate all hierarchical levels together
        merge_clusters = []
        for _, mg in merge_groups.items():
            if len(mg) > 0:
                merge_clusters.extend(mg)

        # sanity check
        checksum = np.sum([len(j) for j in merge_clusters])
        assert checksum == checksum_ref, \
            'checksum fail! not every value is assigned exactly one cluster.'

        # restructure into dicts with roi name, nest id keys for convenience
        def _parse_to_dict(text):
            parts = tuple(text.split('_nid-'))
            return {'roiname': parts[0], 'nid': int(parts[1])}
        merge_clusters = [
            [_parse_to_dict(j) for j in cl] for cl in merge_clusters]

        return merge_clusters

    def _get_merged_polygon(self, cluster):
        """Merge polygons using shapely (Internal).

        Given a single cluster from _get_merge_clusters_from_df(), This creates
        and merges polygons into a single cascaded union. It first dilates the
        polygons by buffer_size pixels to make them overlap, merges them,
        then erodes back by buffer_size to get the merged polygon.

        """
        buffer_size = self.merge_thresh + 3
        nest_polygons = []
        for nestinfo in cluster:
            nest = dict(self.edge_contours[nestinfo['roiname']].loc[
                nestinfo['nid'], :])
            roitop = self.roiinfos[nestinfo['roiname']]['top']
            roileft = self.roiinfos[nestinfo['roiname']]['left']
            coords = _parse_annot_coords(
                nest, x_offset=roileft, y_offset=roitop)
            nest_polygons.append(Polygon(coords).buffer(buffer_size))
        merged_polygon = unary_union(nest_polygons).buffer(-buffer_size)
        return merged_polygon

    def _get_all_merged_polygons(self, merge_clusters, monitorPrefix=''):
        """Merge polygons using shapely (Internal).

        Given a a list of clusters from _get_merge_clusters_from_df(). Creates
        and merges polygons into a single cascaded union. It first dilates the
        polygons by buffer_size pixels to make them overlap, merges them,
        then erodes back by buffer_size to get the merged polygon.

        """
        merged_polygons = []
        for cid, cl in enumerate(merge_clusters):
            self._print2('%s: cluster %d of %d' % (
                monitorPrefix, cid + 1, len(merge_clusters)))
            merged_polygon = self._get_merged_polygon(cluster=cl)
            merged_polygons.append(merged_polygon)
        return merged_polygons

    def _get_coord_str_from_polygon(self, polygon):
        """Parse shapely polygon coordinates into string form (Internal)."""
        coords = np.int32(polygon.exterior.coords.xy)
        coords_x = ','.join([str(j) for j in coords[0, :]])
        coords_y = ','.join([str(j) for j in coords[1, :]])
        return coords_x, coords_y

    def _add_single_merged_edge_contour(self, polygon, group):
        """Add single contour to self.merged_contours (Internal)."""
        idx = self.merged_contours.shape[0]
        self.merged_contours.loc[idx, 'group'] = group
        self.merged_contours.loc[idx, 'color'] = self.GTCodes_df.loc[
            group, 'color']
        self.merged_contours.loc[idx, 'has_holes'] = int(
            polygon.boundary.geom_type == 'MultiLineString')
        coords_x, coords_y = self._get_coord_str_from_polygon(polygon)
        self.merged_contours.loc[idx, 'coords_x'] = coords_x
        self.merged_contours.loc[idx, 'coords_y'] = coords_y

    def _add_merged_edge_contours(
            self, merged_polygons, group, monitorPrefix=''):
        """Add merged polygons to self.merged_contours df (Internal)."""
        for pno, geometry in enumerate(merged_polygons):
            self._print2('%s: contour %d of %d' % (
                monitorPrefix, pno + 1, len(merged_polygons)))
            if geometry.geom_type == 'MultiPolygon':
                for polygon in geometry:
                    self._add_single_merged_edge_contour(polygon, group=group)
            else:
                self._add_single_merged_edge_contour(geometry, group=group)

    def _drop_merged_edge_contours(self, merge_df):
        """Drop edge contours that have already been merged (Internal)."""
        for roiname, _edge_df in self.edge_contours.items():
            nids = []
            for rid in [1, 2]:
                neststr = 'nest%d' % rid
                nids.extend(list(merge_df.loc[merge_df.loc[
                    :, neststr + '-roiname'] == roiname, neststr + '-nid']))
            self.edge_contours[roiname].drop(nids, axis=0, inplace=True)

    def _add_roi_offset_to_contours(self, roi_df, roiname):
        """Add roi offset to coordinates of polygons (Internal)."""
        for idx, annot in roi_df.iterrows():
            coords = np.int32(_parse_annot_coords(dict(annot)))
            coords[:, 0] = coords[:, 0] + self.roiinfos[roiname]['left']
            coords[:, 1] = coords[:, 1] + self.roiinfos[roiname]['top']
            roi_df.loc[idx, 'coords_x'] = ','.join(
                [str(j) for j in coords[:, 0]])
            roi_df.loc[idx, 'coords_y'] = ','.join(
                [str(j) for j in coords[:, 1]])
        return roi_df

    def set_merged_contours(self, monitorPrefix=''):
        """Go through each group and merge contours.

        Sets:
        - self.merged_contours: pandas DataFrame: has the same structure
        as output from get_contours_from_mask().
        """
        for group in self.GTCodes_df.index:

            monitorPrefix = '%s: %s' % (monitorPrefix, group)

            # get pairs of contours to merge
            merge_df = self._get_merge_df(
                group=group, monitorPrefix='%s: _get_merge_df' % monitorPrefix)

            # get clusters of polygons to merge
            merge_clusters = self._get_merge_clusters_from_df(
                merge_df=merge_df,
                monitorPrefix='%s: _get_merge_clusters_from_df' % (
                    monitorPrefix))

            # fetch merged polygons
            merged_polygons = self._get_all_merged_polygons(
                merge_clusters=merge_clusters,
                monitorPrefix='%s: _get_all_merged_polygons' % monitorPrefix)

            # add medged contours to dataframe
            self._add_merged_edge_contours(
                merged_polygons=merged_polygons, group=group,
                monitorPrefix='%s: _add_merged_edge_contours' % monitorPrefix)

            # drop merged edge contours from edge dataframes
            self._print2('%s: _drop_merged_edge_contours' % monitorPrefix)
            self._drop_merged_edge_contours(merge_df=merge_df)

    def get_concatenated_contours(self):
        """Get concatenated contours and overall bounding box.

        Returns
        -------
        - pandas DataFrame: has the same structure as output from
        get_contours_from_mask().

        """
        from pandas import concat

        # concatenate all contours
        all_contours = self.merged_contours.copy()
        for contours_dict in [self.edge_contours, self.ordinary_contours]:
            for roiname, roi_df in contours_dict.items():
                roi_df = self._add_roi_offset_to_contours(
                    roi_df=roi_df, roiname=roiname)
                all_contours = concat(
                    (all_contours, roi_df), axis=0, ignore_index=True)

        # add overall bounding box contour
        idx = all_contours.shape[0]
        top = str(int(np.min([j['top'] for _, j in self.roiinfos.items()])))
        bottom = str(int(
            np.max([j['bottom'] for _, j in self.roiinfos.items()])))
        left = str(int(np.min([j['left'] for _, j in self.roiinfos.items()])))
        right = str(int(
            np.max([j['right'] for _, j in self.roiinfos.items()])))
        all_contours.loc[idx, 'group'] = self.roi_group
        all_contours.loc[idx, 'color'] = self.GTCodes_df.loc['roi', 'color']
        all_contours.loc[idx, 'coords_x'] = ','.join(
            [left, right, right, left, left])
        all_contours.loc[idx, 'coords_y'] = ','.join(
            [top, top, bottom, bottom, top])

        return all_contours

    def run(self):
        """Run full pipeline to get merged contours.

        Returns
        -------
        - pandas DataFrame: has the same structure as output from
        get_contours_from_mask().

        """
        self._print1(
            '\n%s: Set contours from all masks' % self.monitorPrefix)
        self.set_contours_from_all_masks(
            monitorPrefix='%s: set_contours_from_all_masks' %
            self.monitorPrefix)
        self._print1(
            '\n%s: Set ROI bounding boxes' % self.monitorPrefix)
        self.set_roi_bboxes()
        self._print1(
            '\n%s: Set shard ROI edges' % self.monitorPrefix)
        self.set_shared_roi_edges()
        self._print1(
            '\n%s: Set merged contours' % self.monitorPrefix)
        self.set_merged_contours(
            monitorPrefix='%s: set_merged_contours' % self.monitorPrefix)
        self._print1(
            '\n%s: Get concatenated contours' % self.monitorPrefix)
        all_contours = self.get_concatenated_contours()
        if self.discard_nonenclosed_background:
            self._print2(
                '\n%s: Discard nonenclosed background' %
                self.monitorPrefix)
            all_contours = _discard_nonenclosed_background_group(
                all_contours, background_group=self.background_group,
                verbose=self.verbose,
                monitorPrefix='%s: _discard_nonenclosed_background_group' %
                self.monitorPrefix)
        return all_contours



"""
Created on Sun Aug 11 22:50:03 2019.
@author: tageldim
"""
import copy
import os
import sys

import pytest

from histomicstk.annotations_and_masks.annotation_and_mask_utils import (
    get_bboxes_from_slide_annotations, get_image_from_htk_response,
    get_scale_factor_and_appendStr, parse_slide_annotations_into_tables,
    scale_slide_annotations)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../'))
from tests.htk_test_utilities import girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.iteminfo = None
        self.annotations = None


cfg = Cfg()


class TestAnnotAndMaskUtils:
    """Test utilities for annotations and masks."""

    # pytest runs tests in the order they appear in the module
    @pytest.mark.usefixtures('girderClient')  # noqa
    def test_prep(self, girderClient):  # noqa
        cfg.gc = girderClient
        original_iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

        cfg.folder = cfg.gc.post(
            '/folder', data={
                'parentId': original_iteminfo['folderId'],
                'name': 'test-annot-and-mask',
            })

        # copy the item
        cfg.iteminfo = cfg.gc.post(
            '/item/%s/copy' % original_iteminfo['_id'], data={
                'name': 'TCGA-A2-A0YE-01Z.svs',
                'copyAnnotations': True,
                'folderId': cfg.folder['_id'],
            })
        cfg.annotations = cfg.gc.get('/annotation/item/' + cfg.iteminfo['_id'])

    def test_get_image_from_htk_response(self):
        """Test get_image_from_htk_response."""
        getStr = '/item/%s/tiles/region?left=%d&right=%d&top=%d&bottom=%d&encoding=PNG' % (
            cfg.iteminfo['_id'], 59000, 59100, 35000, 35100)
        resp = cfg.gc.get(getStr, jsonResp=False)
        rgb = get_image_from_htk_response(resp)

        assert rgb.shape == (100, 100, 3)

    def test_get_bboxes_from_slide_annotations(self):
        """Test get_bboxes_from_slide_annotations."""
        element_infos = get_bboxes_from_slide_annotations(
            copy.deepcopy(cfg.annotations))

        assert element_infos.shape == (76, 9)
        assert set(element_infos.columns) == {
            'annidx', 'elementidx', 'type', 'group',
            'xmin', 'xmax', 'ymin', 'ymax', 'bbox_area'}

    def test_parse_slide_annotations_into_tables(self):
        """Test parse_slide_annotations_into_tables."""
        annotation_infos, element_infos = parse_slide_annotations_into_tables(
            copy.deepcopy(cfg.annotations))

        assert set(annotation_infos.columns) == {
            'annotation_girder_id', '_modelType', '_version',
            'itemId', 'created', 'creatorId',
            'public', 'updated', 'updatedId',
            'groups', 'element_count', 'element_details',
        }
        assert set(element_infos.columns) == {
            'annidx', 'annotation_girder_id',
            'elementidx', 'element_girder_id',
            'type', 'group', 'label', 'color',
            'xmin', 'xmax', 'ymin', 'ymax', 'bbox_area',
            'coords_x', 'coords_y',
        }
        assert set(element_infos.loc[:, 'type']) == {'polyline', 'rectangle'}

    def test_scale_slide_annotations(self):
        """Test scale_slide_annotations."""
        for sf in (0.5, 1.0):
            modified = scale_slide_annotations(
                copy.deepcopy(cfg.annotations), sf=sf)
            assert modified[0]['annotation']['elements'][0]['center'] == [
                int(sf * j) for j in
                cfg.annotations[0]['annotation']['elements'][0]['center']]

    def test_get_scale_factor_and_appendStr(self):
        """Test get_scale_factor_and_appendStr."""
        in_out = [
            [(0.2, None), (1.2525, '&mm_x=0.00020000&mm_y=0.00020000')],
            [(None, 10.), (0.25, '&magnification=10.00000000')],
            [(None, None), (1.0, '')],
        ]
        for (MPP, MAG), (sftrue, apstr) in in_out:
            sf, appendStr = get_scale_factor_and_appendStr(
                gc=cfg.gc, slide_id=cfg.iteminfo['_id'], MPP=MPP, MAG=MAG)
            assert sf == sftrue
            assert appendStr == apstr



import numpy as np

import histomicstk.utils as htk_utls


def reg_edge(im_input, im_phi, well='double', sigma=1.5, dt=1.0, mu=0.2,
             lambda_=1, alpha=-3, epsilon=1.5, iter=100):
    """Distance-regularized edge-based level sets.

    Distance-regularization is used in this edge-based level set implementation
    to avoid numerical problems requiring costly re-initialization. Provides
    cost terms for boundary length, area, and regularization of the level set
    function. Foreground objects are assumed to have larger intensity values
    than background.

    Parameters
    ----------
    im_input : array_like
        A floating-point intensity image.
    im_phi : array_like
        A floating-point initialization of the level-set image. Interior values
        are set to -c0, and exterior values set to c0, where c0 > 0.
    well : string
        Choice of well function for regularization. Can be set to either
        'single' or 'double' for single-well or double-well regularization, or
        any other value for no regularization. Default value = 'double'.
    sigma : double
        Standard deviation of smoothing filter for input image im_input.
    dt : double
        Time step for evolving im_phi. Default value = 1.0.
    mu : double
        Regularization weight for energy function. Default value = 0.2.
    lambda_ : double
        Boundary length weight for energy function. Default value = 1.0.
    alpha : double
        Area weight for energy function. A negative value is used to seed the
        interior of the foreground objects and then evolve the boundary
        outwards. A positive value assumes that the boundary begins outside the
        foreground objects and collapses to their high-gradient edges.
        Default value = -3.
    epsilon: double
        Coefficient used to smooth the Dirac and Heaviside functions. Default
        value = 1.5.
    iter: double
        Number of iterations to evolve curve level set function over. Default
        value = 100.

    Returns
    -------
    im_phi : array_like
        An intensity image where the zero level set defines object boundaries.
        Can be further processed with fast marching methods or other to obtain
        smooth boundaries, or simply thresholded to define the object mask.

    See Also
    --------
    histomicstk.segmentation.nuclear.gaussian_voting

    References
    ----------
    .. [#] C. Li, C. Xu, C. Gui, M.D. Fox, "Distance Regularized Level Set
       Evolution and Its Application to Image Segmentation," in IEEE
       Transactions on Image Processing, vol.19,no.12,pp.3243-54, 2010.

    """
    import scipy.ndimage as ndi

    # smoothed gradient of input image
    sI = ndi.gaussian_filter(im_input, sigma, mode='constant', cval=0)
    dsI = np.gradient(sI)
    G = 1 / (1 + dsI[0]**2 + dsI[1]**2)
    dG = np.gradient(G)

    # perform regularized level-set evolutions with time step dt
    for i in range(iter):

        # fix boundary conditions
        im_phi = neumann_bounds(im_phi)

        # calculate gradient of level set image
        dPhi = np.gradient(im_phi)
        mPhi = (dPhi[0]**2 + dPhi[1]**2)**0.5  # gradient magnitude
        Curve = np.gradient(dPhi[0] / (mPhi + 1e-10))[0] + \
            np.gradient(dPhi[1] / (mPhi + 1e-10))[1]  # divergence

        # build regularization function
        if well == 'single':
            Reg = single_well(im_phi, Curve)
        elif well == 'double':
            Reg = double_well(im_phi, dPhi, mPhi, Curve, i)
        else:
            Reg = np.zeros(im_phi.shape)

        # area and boundary-length energy function terms
        iPhi = impulse(im_phi, epsilon)
        Area = iPhi * G
        Edge = iPhi * (dG[0] * (dPhi[0] / (mPhi + 1e-10)) +
                       dG[1] * (dPhi[1] / (mPhi + 1e-10))) + iPhi * G * Curve

        # evolve level-set function
        im_phi = im_phi + dt * (mu * Reg + lambda_ * Edge + alpha * Area)

    # return evolved level-set function following iterations
    return im_phi


def initialize(Mask, c0=2):
    # initialize scaled binary-step image
    Phi0 = np.zeros(Mask.shape)
    Phi0[Mask > 0] = -c0
    Phi0[Mask == 0] = c0
    return Phi0


def single_well(Phi, Curve):
    # Single-well potential function
    return 4 * htk_utls.del2(Phi) - Curve


def double_well(Phi, dPhi, mPhi, Curve, i):
    # Double-well potential function
    SmallMask = (mPhi <= 1) & (mPhi >= 0)
    LargeMask = (mPhi > 1)
    P = SmallMask * np.sin(2 * np.pi * mPhi) / \
        (2 * np.pi) + LargeMask * (mPhi - 1)
    dP = ((P != 0) * P + (P == 0)) / ((mPhi != 0) * mPhi + (mPhi == 0))
    Well = np.gradient(dP * dPhi[0] - dPhi[0])[0] + \
        np.gradient(dP * dPhi[1] - dPhi[1])[1] + 4 * htk_utls.del2(Phi)
    return Well


def impulse(X, Epsilon):
    # Smooth dirac delta function.

    # calculate smoothed impulse everywhere
    Xout = (1 + np.cos(np.pi * X / Epsilon)) / (2 * Epsilon)

    # zero out values |x| > Epsilon
    Xout[np.absolute(X) > Epsilon] = 0

    return Xout


def neumann_bounds(Phi):
    # Transform input to enforce Neumann boundary conditions.

    # copy input
    PhiOut = Phi

    # capture image size
    m = Phi.shape[0]
    n = Phi.shape[1]

    # deal with corners
    PhiOut[0, 0] = PhiOut[2, 2]
    PhiOut[0, n - 1] = PhiOut[0, -3]
    PhiOut[m - 1, 0] = PhiOut[-3, 2]
    PhiOut[m - 1, n - 1] = PhiOut[-3, -3]

    # deal with edges
    PhiOut[0, 1:-1] = PhiOut[2, 1:-1]
    PhiOut[m - 1, 1:-1] = PhiOut[m - 3, 1:-1]
    PhiOut[1:-1, 0] = PhiOut[1:-1, 2]
    PhiOut[1:-1, n - 1] = PhiOut[1:-1, n - 3]

    return PhiOut



"""Module to run specific workflows on groups of slides.

Note that this is separate from the server-side way of doing things.
This is just a convenience utility for developers and machine learning
engineers who would like to run the same function over all slides
in a girder directory.

"""
